/*
 * Final Project: PomodorIO - The Digi-Physical Time Management Method!
 * 
 * We created PomodorIO, inspired by Pomodoro Technique, which is a time management method developed by Francesco Cirillo in the late 1980s.
 * With PomodorIO you can set manage your time and tasks easily, and feel more productive!
 * 
 * For more information about the project, visit PomodorIO's Instructable at:
 * https://www.instructables.com/-PomodorIO-the-Digi-Physical-Time-Management-Metho/
 * 
 * Created by:
 * Tomer Ben-Gigi, 206198772
 * Lior Reytan, 204326607
 */

#include <ESP8266_Lib.h>
#include <BlynkSimpleShieldEsp8266.h>
#include <Adafruit_CircuitPlayground.h>
#include <TimeLib.h>
#include <WidgetRTC.h>

#define EspSerial Serial1
#define ESP8266_BAUD 115200
#define BLYNK_PRINT SerialUSB
#define BLYNK_DEBUG SerialUSB
#define BLYNK_UPDATE_INTERVAL 100

#define SETUP_LED_COLOR 0xEC2D01       // Tomato red color
#define WORK_LED_COLOR 0xFF0000        // Red color
#define SHORT_BREAK_LED_COLOR 0xBFFF00 // Lime color
#define LONG_BREAK_LED_COLOR 0x009000  // Islamic green color
#define PAUSE_LED_COLOR 0xFFA500       // Orange color
#define NOTIFICATION_LED 13            // The LED used for Integromat notifications
#define NUM_OF_SHORT_BREAKS 3          // Number of short breaks before the long break, according to the Pomodoro paradigm
#define NOTIFICATION_DURATION 0.5
#define DELAY_BETWEEN_POMODOROS 0.25
#define TAP_THRESHOLD 120

// Enum - Finite State Machine
enum StateEnum
{
  SETUP,
  WORK,
  NOTIFICATION,
  PAUSE,
  BREAK
};

// *** GLOBALS ***

// Connectivity
char auth[] = "xPuwmPrcjB4B27T0pti2VOpY4dt8JZ8H";
char ssid[] = "ReyNet_UP";
char pass[] = "VE3974849re$";
ESP8266 wifi(&EspSerial);

// State
int state = SETUP;
int stateBeforePausing; // Used for correct resuming after exiting PAUSE state

// Time settings
int pomodoroTime;     // The work time (pomodoro time) in minutes, determined in the Blynk app
int shortBreakTime;   // The short break time in minutes, determined in the Blynk app
int longBreakTime;    // The long break time in minutes, determined in the Blynk app
int currentBreakTime; // Used for changing between the short and long break times

int breaksCounter;         // Used for counting the short breaks until the long break
bool isStartButtonOn;      // Used to determine whether the start button was pressed in the Blynk app
bool shouldNotify = false; // Used to determine whether a new notification should be notified
int currentLedColor;       // Used to determine the current color of the LEDs according to the current state

// Time-related globals
unsigned long targetMillisTime;  // Used for WORK and BREAK states timers (in milliseconds)
unsigned long onPauseMillisTime; // The time when the slide switch turned on (in milliseconds)
BlynkTimer timer;                // Used for wrapping virtual write commands

// Pomodoros counter
int taskPomodorosCounter = 1; // Used for counting the number of pomodoros taken for the current task
int currentTaskRow = 3;       // Indicates the current task row in the Google Sheets file

// Notification LED
int notifLed_Brightness = 0;
int notifLed_FadeAmount = 5;
bool isNotifLedOn;

WidgetTerminal terminal(V10); // Attach virtual serial terminal to Virtual Pin V10

void setup()
{
  CircuitPlayground.begin();
  SerialUSB.begin(9600);
  EspSerial.begin(ESP8266_BAUD);

  // For the double-tap gesture
  CircuitPlayground.setAccelRange(LIS3DH_RANGE_2_G);
  CircuitPlayground.setAccelTap(2, TAP_THRESHOLD);
  attachInterrupt(digitalPinToInterrupt(CPLAY_LIS3DH_INTERRUPT), onDoubleTapped, FALLING);

  delay(10);
  Blynk.begin(auth, wifi, ssid, pass);
}

void loop()
{
  // Initiates Blynk-related features
  Blynk.run();
  timer.run();

  // Any state transitions
  state = anyStateTransitions();

  // Displays the notification LED (D13) if Integromat notification has been received
  displayNotificationLed();

  // Switches between the different states
  switch (state)
  {
    case SETUP:
      state = setupState();
      break;

    case WORK:
      state = workState();
      break;

    case NOTIFICATION:
      state = notificationState();
      break;

    case PAUSE:
      state = pauseState();
      break;

    case BREAK:
      state = breakState();
      break;
  }
}

// Defines the transition from any possible state
int anyStateTransitions()
{
  // If the start button in the Blynk app is not pressed OR the CPX's left button pressed
  // - return to SETUP state
  if (!isStartButtonOn || CircuitPlayground.leftButton())
  {
    return SETUP;
  }

  // If the slide switch is turned on AND we are currently not in PAUSE state
  // - return to PAUSE state
  if (CircuitPlayground.slideSwitch() && state != PAUSE)
  {
    onPauseMillisTime = millis();
    stateBeforePausing = state;
    Serial.println("* PAUSE STATE *");

    return PAUSE;
  }

  // If the slide switch is turned off AND we are currently in PAUSE state
  // - return to the state we were before pausing
  if (!CircuitPlayground.slideSwitch() && state == PAUSE)
  {
    targetMillisTime += millis() - onPauseMillisTime;

    return stateBeforePausing;
  }

  return state;
}

// *** BLYNK EVENTS ***

// Triggred when the CPX board connects to the Blynk app
BLYNK_CONNECTED()
{
  // Synchronize settings virtual pins on connection
  Blynk.syncVirtual(V0, V1, V2, V4);
  Serial.println("Blynk is connected");
}

// Blynk app widget - Pomodoro time
BLYNK_WRITE(V0)
{
  if (state == SETUP)
  {
    pomodoroTime = param.asInt();
  }
}

// Blynk app widget - Short break time
BLYNK_WRITE(V1)
{
  if (state == SETUP)
  {
    shortBreakTime = param.asInt();
  }
}

// Blynk app widget - Long break time
BLYNK_WRITE(V2)
{
  if (state == SETUP)
  {
    longBreakTime = param.asInt();
  }
}

// Blynk app widget - Start button
BLYNK_WRITE(V4)
{
  isStartButtonOn = param.asInt();
}

// Blynk - Integromat notification
BLYNK_WRITE(V5)
{
  if (param.asInt() == 1)
  {
    timer.setTimeout(1000, onNotificationReceived);
  }
}

// *** FSM STATES FUNCTIONS ***

// Defines the behavior and transitions of SETUP state
int setupState()
{
  isNotifLedOn = false;
  currentLedColor = SETUP_LED_COLOR;
  displaySetupStateAnimation();

  if (isStartButtonOn && !CircuitPlayground.leftButton())
  {
    // Start a new pomodoro.
    targetMillisTime = millis() + (60000 * pomodoroTime); // 1 Minute = 60000 Milliseconds
    Serial.println("* WORK STATE *");

    return WORK;
  }

  return SETUP;
}

// Defines the behavior and transitions of WORK state
int workState()
{
  currentLedColor = WORK_LED_COLOR;
  if (shouldNotify)
  {
    shouldNotify = false;

    return NOTIFICATION;
  }

  if (millis() > targetMillisTime || CircuitPlayground.rightButton())
  {
    taskPomodorosCounter++;
    Serial.println("* BREAK STATE *");
    if (breaksCounter == NUM_OF_SHORT_BREAKS)
    {
      // Long break
      currentBreakTime = longBreakTime;
      breaksCounter = 0;
      currentLedColor = LONG_BREAK_LED_COLOR;
    }
    else
    {
      // Short break
      currentBreakTime = shortBreakTime;
      breaksCounter++;
      currentLedColor = SHORT_BREAK_LED_COLOR;
    }

    targetMillisTime = millis() + (60000 * currentBreakTime); // 1 Minute = 60000 Milliseconds
    delay(DELAY_BETWEEN_POMODOROS * 1000);

    return BREAK;
  }

  displayTimerLeds(false, pomodoroTime);

  return WORK;
}

// Defines the behavior and transitions of NOTIFICATION state
int notificationState()
{
  CircuitPlayground.playTone(440, NOTIFICATION_DURATION * 1000);
  isNotifLedOn = true;

  return WORK;
}

// Defines the behavior and transitions of PAUSE state
int pauseState()
{
  CircuitPlayground.clearPixels();
  currentLedColor = PAUSE_LED_COLOR;

  for (int i = 0; i < 10; ++i)
  {
    CircuitPlayground.setPixelColor(i, currentLedColor);
  }

  return PAUSE;
}

// Defines the behavior and transitions of BREAK state
int breakState()
{
  // Pause switch
  if (CircuitPlayground.slideSwitch())
  {
    onPauseMillisTime = millis();
    stateBeforePausing = BREAK;
    Serial.println("* PAUSE STATE *");

    return PAUSE;
  }

  if (millis() > targetMillisTime || CircuitPlayground.rightButton())
  {
    Serial.println("* SETUP STATE *");

    delay(DELAY_BETWEEN_POMODOROS * 1000);

    return SETUP;
  }

  displayTimerLeds(true, currentBreakTime);

  return BREAK;
}

// *** HELPER FUNCTIONS ***

// Gets a boolean that determines if the LED lights are increasing and the duration of current state (WORK / BREAK)
// and displays the LED timer visualization accordingly
void displayTimerLeds(bool isIncreasingLeds, int duration)
{
  float lightIntervalInMinutes = (duration / 10.0);
  unsigned long remainingTime = targetMillisTime - millis();
  float remainingTimeInMinutes = remainingTime / 60000.0;
  float runningTimeInMinutes = duration - remainingTimeInMinutes;
  int ledsToLight = runningTimeInMinutes / lightIntervalInMinutes;

  ledsToLight = isIncreasingLeds ? ledsToLight : 10 - ledsToLight;
  CircuitPlayground.clearPixels();
  for (int i = 0; i < ledsToLight; i++)
  {
    CircuitPlayground.setPixelColor(i, currentLedColor);
  }
}

// Event handler for double-tap gesture
void onDoubleTapped(void)
{
  CircuitPlayground.playTone(1000, NOTIFICATION_DURATION * 1000);
  Serial.println("TAP!");
  if (state == WORK)
  {
    timer.setTimeout(1000, onTaskCompleted);
  }
}

// Event handler for Integromat notification receiving
void onNotificationReceived()
{
  shouldNotify = true;
  Blynk.virtualWrite(5, 0);
}

// Event handler called by the double-tap gesture event to write the relevant data to V6
// (and from there to the Google Sheets file thru Integromat)
void onTaskCompleted()
{
  // Send task data
  Serial.println("sending data");
  Serial.println(currentTaskRow);
  Serial.println(taskPomodorosCounter);
  Blynk.virtualWrite(6, currentTaskRow, taskPomodorosCounter, shortBreakTime, longBreakTime, pomodoroTime);
  terminal.println(currentTaskRow);
  terminal.flush();

  // Prepare for next task
  currentTaskRow++;
  taskPomodorosCounter = 1;
}

// Displays a spectacular animation for SETUP state
void displaySetupStateAnimation()
{
  int led1 = (millis() / 100) % 10;
  int led2 = (led1 + 1) % 10;
  int led3 = (led2 + 1) % 10;
  
  CircuitPlayground.clearPixels();
  CircuitPlayground.setPixelColor(led1, currentLedColor); 
  CircuitPlayground.setPixelColor(led2, currentLedColor); 
  CircuitPlayground.setPixelColor(led3, currentLedColor); 
}

// Displays the notification LED (D13) after receiving Integromat notification
void displayNotificationLed()
{
  if (isNotifLedOn)
  {
    int delay = (millis() / 2000) % 2;
  
    analogWrite(NOTIFICATION_LED, notifLed_Brightness);
    notifLed_Brightness = notifLed_Brightness + notifLed_FadeAmount * delay;
    if (notifLed_Brightness <= 0 || notifLed_Brightness >= 255)
    {
      notifLed_FadeAmount = -notifLed_FadeAmount;
    }
  }
  else
  {
    analogWrite(NOTIFICATION_LED, 0);
  }
}